/*
 * Copyright (c) 2012 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is confidential property of Nordic Semiconductor. The use,
 * copying, transfer or disclosure of such information is prohibited except by express written
 * agreement with Nordic Semiconductor.
 *
 */

/*  This module is a modified version of the Cycling Speed and Cadence Service Client provided with the SDK from
 *  Nordic Semiconductors. This module is not initialized with a single client structure, instead
 *  it is initialized with an array of Client Structures. Therefore it is possible to handle more
 *  than one hrs client with this module. The handling of the array is done automatically by this
 *  module.
 */

/**@cond To Make Doxygen skip documentation generation for this file.
 * @{
 */

#include <stddef.h>
#include <stdint.h>
#include <string.h>
#include "ble_csc_c.h"
#include "ble_db_discovery.h"
#include "ble_types.h"
#include "ble_srv_common.h"
#include "nordic_common.h"
#include "nrf_error.h"
#include "ble_gattc.h"
#include "app_util.h"
//#include "app_trace.h"

//#define LOG                    app_trace_log         /**< Debug logger macro that will be used in this file to do logging of important information over UART. */
#define LOG(...)

#define CSCM_FLAG_WHEEL_PRESENT  (0x01 << 0)           /**< Bit mask used to extract the presence of Wheel Revolution Data. */
#define CSCM_FLAG_CRANK_PRESENT  (0x01 << 1)           /**< Bit mask used to extract the presence of Crank Revolution Data. */

#define CSCF_FLAG_WHEEL_SUPP     (0x01 << 0)         /**< Bit mask to extract if Wheel Revolution Data is supported. */
#define CSCF_FLAG_CRANK_SUPP     (0x01 << 1)         /**< Bit mask to extract if Crank Revolution Data is supported. */
#define CSCF_FLAG_MULT_SENS_SUPP (0x01 << 2)         /**< Bit mask to extract if Multiple Sensor Locations is supported. */

#define TX_BUFFER_MASK         0x07                  /**< TX Buffer mask, must be a mask of continuous zeroes, followed by continuous sequence of ones: 000...111. */
#define TX_BUFFER_SIZE         (TX_BUFFER_MASK + 1)  /**< Size of send buffer, which is 1 higher than the mask. */

#define WRITE_MESSAGE_LENGTH   BLE_CCCD_VALUE_LEN    /**< Length of the write message for CCCD. */

typedef enum
{
    READ_REQ,  /**< Type identifying that this tx_message is a read request. */
    WRITE_REQ  /**< Type identifying that this tx_message is a write request. */
} tx_request_t;

/**@brief Structure for writing a message to the peer, i.e. CCCD.
 */
typedef struct
{
    uint8_t                  gattc_value[WRITE_MESSAGE_LENGTH];  /**< The message to write. */
    ble_gattc_write_params_t gattc_params;                       /**< GATTC parameters for this message. */
} write_params_t;

/**@brief Structure for holding data to be transmitted to the connected central.
 */
typedef struct
{
    uint16_t     conn_handle;  /**< Connection handle to be used when transmitting this message. */
    tx_request_t type;         /**< Type of this message, i.e. read or write message. */
    union
    {
        uint16_t       read_handle;  /**< Read request message. */
        write_params_t write_req;    /**< Write request message. */
    } req;
} tx_message_t;


static ble_csc_c_t * mp_ble_csc_c_start;           /**< Pointer to the first instance of the CSC Client module */
static uint8_t       m_csc_c_numof;                /**< Number of structs available */
static tx_message_t  m_tx_buffer[TX_BUFFER_SIZE];  /**< Transmit buffer for messages to be transmitted to the central. */
static uint32_t      m_tx_insert_index = 0;        /**< Current index in the transmit buffer where the next message should be inserted. */
static uint32_t      m_tx_index = 0;               /**< Current index in the transmit buffer from where the next message to be transmitted resides. */

/**@brief Function for passing any pending request from the buffer to the stack.
 */
static void tx_buffer_process(void)
{
    if (m_tx_index != m_tx_insert_index)
    {
        uint32_t err_code;

        if (m_tx_buffer[m_tx_index].type == READ_REQ)
        {
            err_code = sd_ble_gattc_read(m_tx_buffer[m_tx_index].conn_handle,
                                         m_tx_buffer[m_tx_index].req.read_handle,
                                         0);
        }
        else
        {
            err_code = sd_ble_gattc_write(m_tx_buffer[m_tx_index].conn_handle,
                                          &m_tx_buffer[m_tx_index].req.write_req.gattc_params);
        }
        if (err_code == NRF_SUCCESS)
        {
            LOG("[HRS_C]: SD Read/Write API returns Success..\r\n");
            m_tx_index++;
            m_tx_index &= TX_BUFFER_MASK;
        }
        else
        {
            LOG("[HRS_C]: SD Read/Write API returns error. This message sending will be "
                "attempted again..\r\n");
        }
    }
}


/**@brief     Function for handling write response events.
 *
 * @param[in] p_ble_csc_c Pointer to the Cycling Speed and Cadence Client structure.
 * @param[in] p_ble_evt   Pointer to the BLE event received.
 */
static void on_write_rsp(ble_csc_c_t * p_ble_csc_c, const ble_evt_t * p_ble_evt)
{
    // Check if there is any message to be sent across to the peer and send it.
    tx_buffer_process();
}


/**@brief     Function for handling read response events.
 *
 * @details   This function will validate the read response and raise the appropriate
 *            event to the application.
 *
 * @param[in] p_csc_c   Pointer to the Cycling Speed and Cadence Client Structure.
 * @param[in] p_ble_evt Pointer to the SoftDevice event.
 */
static void on_read_rsp(ble_csc_c_t * p_csc_c, const ble_evt_t * p_ble_evt)
{
    const ble_gattc_evt_read_rsp_t * p_response;

    p_response = &p_ble_evt->evt.gattc_evt.params.read_rsp;

    if (p_response->handle == p_csc_c->cscf_handle)
    {
        ble_csc_c_evt_t evt;
        uint16_t feature;

        evt.evt_type = BLE_CSC_C_EVT_CSCF_READ_RESP;

        feature = uint16_decode(&p_response->data[0]);
        evt.params.cscf.wheel_revo_supported = (feature & CSCF_FLAG_WHEEL_SUPP ? 1 : 0);
        evt.params.cscf.crank_revo_supported = (feature & CSCF_FLAG_CRANK_SUPP ? 1 : 0);
        evt.params.cscf.multiple_sensor_loc  = (feature & CSCF_FLAG_MULT_SENS_SUPP ? 1 : 0);

        p_csc_c->evt_handler(p_csc_c, &evt);
    }
    // Check if there is any buffered transmissions and send them.
    tx_buffer_process();
}


/**@brief     Function for handling Handle Value Notification received from the SoftDevice.
 *
 * @details   This function will uses the Handle Value Notification received from the SoftDevice
 *            and checks if it is a notification of the heart rate measurement from the peer. If
 *            it is, this function will decode the heart rate measurement and send it to the
 *            application.
 *
 * @param[in] p_ble_csc_c Pointer to the Cycling Speed and Cadence Client structure.
 * @param[in] p_ble_evt   Pointer to the BLE event received.
 */
static void on_hvx(ble_csc_c_t * p_ble_csc_c, const ble_evt_t * p_ble_evt)
{
    // Check if this is a cycling speed and cadence notification.
    if (p_ble_evt->evt.gattc_evt.params.hvx.handle == p_ble_csc_c->cscm_handle)
    {
        ble_csc_c_evt_t ble_csc_c_evt;
        uint_fast8_t    flags;
        uint32_t        index = 0;

        memset(&ble_csc_c_evt, 0, sizeof(ble_csc_c_evt_t));
        ble_csc_c_evt.evt_type = BLE_CSC_C_EVT_CSCM_NOTIFICATION;

        flags = p_ble_evt->evt.gattc_evt.params.hvx.data[index++];
        if ((flags & CSCM_FLAG_WHEEL_PRESENT))
        {
            ble_csc_c_evt.params.cscm.wheel_present = 1;
            ble_csc_c_evt.params.cscm.wheel_revolutions =
                uint32_decode(&p_ble_evt->evt.gattc_evt.params.hvx.data[index]);
            index += 4;
            ble_csc_c_evt.params.cscm.wheel_event_time =
                uint16_decode(&p_ble_evt->evt.gattc_evt.params.hvx.data[index]);
            index += 2;
        }
        if ((flags & CSCM_FLAG_CRANK_PRESENT))
        {
            ble_csc_c_evt.params.cscm.crank_present = 1;
            ble_csc_c_evt.params.cscm.crank_revolutions =
                uint16_decode(&p_ble_evt->evt.gattc_evt.params.hvx.data[index]);
            index += 2;
            ble_csc_c_evt.params.cscm.crank_event_time =
                uint16_decode(&p_ble_evt->evt.gattc_evt.params.hvx.data[index]);
            index += 2;
        }

        p_ble_csc_c->evt_handler(p_ble_csc_c, &ble_csc_c_evt);
    }
}


/**@brief     Function for handling events from the database discovery module.
 *
 * @details   This function will handle an event from the database discovery module, and determine
 *            if it relates to the discovery of heart rate service at the peer. If so, it will
 *            call the application's event handler indicating that the cycling speed and cadence service has been
 *            discovered at the peer. It also populates the event with the service related
 *            information before providing it to the application.
 *
 * @param[in] p_evt Pointer to the event received from the database discovery module.
 *
 */
static void db_discover_evt_handler(ble_db_discovery_evt_t * p_evt)
{
    // Check if the Cycling Speed and Cadence Service was discovered.
    if (p_evt->evt_type == BLE_DB_DISCOVERY_COMPLETE &&
        p_evt->params.discovered_db.srv_uuid.uuid == BLE_UUID_CYCLING_SPEED_AND_CADENCE &&
        p_evt->params.discovered_db.srv_uuid.type == BLE_UUID_TYPE_BLE)
    {
        ble_csc_c_t* p_ble_csc_c = NULL;
        //search the first empty cycling speed and cadence client struct
        for (uint8_t i = 0; i < m_csc_c_numof; i++)
        {
            if (mp_ble_csc_c_start[i].conn_handle == BLE_CONN_HANDLE_INVALID)
            {
                p_ble_csc_c = &mp_ble_csc_c_start[i];
                break;
            }
        }
        if (p_ble_csc_c == NULL)
            return;

        p_ble_csc_c->conn_handle = p_evt->conn_handle;

        // Find the CCCD Handle of the Cycling Speed and Cadence Measurement characteristic.
        uint32_t i;

        for (i = 0; i < p_evt->params.discovered_db.char_count; i++)
        {
            if (p_evt->params.discovered_db.charateristics[i].characteristic.uuid.uuid ==
                BLE_UUID_CSC_MEASUREMENT_CHAR)
            {
                // Found Cycling Speed and Cadence measurement characteristic. Store CCCD handle.
                p_ble_csc_c->cscm_cccd_handle =
                    p_evt->params.discovered_db.charateristics[i].cccd_handle;
                p_ble_csc_c->cscm_handle      =
                    p_evt->params.discovered_db.charateristics[i].characteristic.handle_value;
            }
            else if (p_evt->params.discovered_db.charateristics[i].characteristic.uuid.uuid ==
                BLE_UUID_CSC_FEATURE_CHAR)
            {
                // Found Cycling Speed and Cadence feature characteristic.
                p_ble_csc_c->cscf_handle      =
                    p_evt->params.discovered_db.charateristics[i].characteristic.handle_value;
            }
        }

        LOG("[HRS_C]: Cycling Speed and Cadence Service discovered at peer.\r\n");

        ble_csc_c_evt_t evt;

        evt.evt_type = BLE_CSC_C_EVT_DISCOVERY_COMPLETE;

        p_ble_csc_c->evt_handler(p_ble_csc_c, &evt);
    }
}


uint32_t ble_csc_c_init(ble_csc_c_t * p_ble_csc_c, uint8_t num_csc_c, ble_csc_c_init_t * p_ble_csc_c_init)
{
    if ((p_ble_csc_c == NULL) || (p_ble_csc_c_init == NULL))
    {
        return NRF_ERROR_NULL;
    }
    if (num_csc_c == 0)
    {
        return NRF_ERROR_DATA_SIZE;
    }

    ble_uuid_t csc_uuid;

    csc_uuid.type = BLE_UUID_TYPE_BLE;
    csc_uuid.uuid = BLE_UUID_CYCLING_SPEED_AND_CADENCE;

    mp_ble_csc_c_start = p_ble_csc_c;
    m_csc_c_numof      = num_csc_c;

    for (uint8_t i = 0; i < m_csc_c_numof; i++)
    {
        mp_ble_csc_c_start[i].conn_handle      = BLE_CONN_HANDLE_INVALID;
        mp_ble_csc_c_start[i].cscm_cccd_handle = BLE_GATT_HANDLE_INVALID;
        mp_ble_csc_c_start[i].cscm_handle      = BLE_GATT_HANDLE_INVALID;
        mp_ble_csc_c_start[i].cscf_handle      = BLE_GATT_HANDLE_INVALID;
        mp_ble_csc_c_start[i].evt_handler      = p_ble_csc_c_init->evt_handler;
    }

    return ble_db_discovery_evt_register(&csc_uuid,
                                         db_discover_evt_handler);
}


void ble_csc_c_on_ble_evt(const ble_evt_t * p_ble_evt)
{
    ble_csc_c_t * p_ble_csc_c = NULL;

    if ((p_ble_evt == NULL) || (mp_ble_csc_c_start == NULL))
    {
        return;
    }
    for (uint_fast8_t i = 0; i < m_csc_c_numof; i++)
    {
        if (mp_ble_csc_c_start[i].conn_handle == p_ble_evt->evt.gap_evt.conn_handle)
        {
            p_ble_csc_c = &mp_ble_csc_c_start[i];
        }
    }
    if (p_ble_csc_c == NULL)
        return;

    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_DISCONNECTED:
            p_ble_csc_c->conn_handle      = BLE_CONN_HANDLE_INVALID;
            p_ble_csc_c->cscm_cccd_handle = BLE_GATT_HANDLE_INVALID;
            p_ble_csc_c->cscm_handle      = BLE_GATT_HANDLE_INVALID;
            p_ble_csc_c->cscf_handle      = BLE_GATT_HANDLE_INVALID;
            break;

        case BLE_GATTC_EVT_HVX:
            on_hvx(p_ble_csc_c, p_ble_evt);
            break;

        case BLE_GATTC_EVT_WRITE_RSP:
            on_write_rsp(p_ble_csc_c, p_ble_evt);
            break;

        case BLE_GATTC_EVT_READ_RSP:
            on_read_rsp(p_ble_csc_c, p_ble_evt);
            break;

        default:
            break;
    }

    tx_buffer_process();
}


/**@brief Function for creating a message for writing to the CCCD.
 */
static uint32_t cccd_configure(uint16_t conn_handle, uint16_t handle_cccd, bool enable)
{
    LOG("[HRS_C]: Configuring CCCD. CCCD Handle = %d, Connection Handle = %d\r\n",
        handle_cccd,conn_handle);

    tx_message_t * p_msg;
    uint16_t       cccd_val = enable ? BLE_GATT_HVX_NOTIFICATION : 0;

    p_msg              = &m_tx_buffer[m_tx_insert_index++];
    m_tx_insert_index &= TX_BUFFER_MASK;

    p_msg->req.write_req.gattc_params.handle   = handle_cccd;
    p_msg->req.write_req.gattc_params.len      = WRITE_MESSAGE_LENGTH;
    p_msg->req.write_req.gattc_params.p_value  = p_msg->req.write_req.gattc_value;
    p_msg->req.write_req.gattc_params.offset   = 0;
    p_msg->req.write_req.gattc_params.write_op = BLE_GATT_OP_WRITE_REQ;
    p_msg->req.write_req.gattc_value[0]        = LSB_16(cccd_val);
    p_msg->req.write_req.gattc_value[1]        = MSB_16(cccd_val);
    p_msg->conn_handle                         = conn_handle;
    p_msg->type                                = WRITE_REQ;

    tx_buffer_process();
    return NRF_SUCCESS;
}


uint32_t ble_csc_c_cscm_notif_enable(uint16_t conn_handle, bool enable)
{
    ble_csc_c_t * p_ble_csc_c = NULL;

    for (uint8_t i = 0; i < m_csc_c_numof; i++)
    {
        if (mp_ble_csc_c_start[i].conn_handle == conn_handle)
        {
            p_ble_csc_c = &mp_ble_csc_c_start[i];
        }
    }
    if (p_ble_csc_c == NULL)
        return NRF_ERROR_NOT_FOUND;

    return cccd_configure(conn_handle, p_ble_csc_c->cscm_cccd_handle, enable);
}


uint32_t ble_csc_c_cscf_read(uint16_t conn_handle)
{
    tx_message_t * msg;
    ble_csc_c_t * p_ble_csc_c = NULL;

    for (uint8_t i = 0; i < m_csc_c_numof; i++)
    {
        if (mp_ble_csc_c_start[i].conn_handle == conn_handle)
        {
            p_ble_csc_c = &mp_ble_csc_c_start[i];
        }
    }
    if (p_ble_csc_c == NULL)
    {
        return NRF_ERROR_NOT_FOUND;
    }

    msg                  = &m_tx_buffer[m_tx_insert_index++];
    m_tx_insert_index   &= TX_BUFFER_MASK;

    msg->req.read_handle = p_ble_csc_c->cscf_handle;
    msg->conn_handle     = p_ble_csc_c->conn_handle;
    msg->type            = READ_REQ;

    tx_buffer_process();
    return NRF_SUCCESS;
}

/** @}
 *  @endcond
 */
